<?php

namespace Drupal\sqrl\Plugin\SqrlAction;

use Drupal\Core\Entity\EntityStorageException;
use Drupal\sqrl\Entity\Identity;
use Drupal\sqrl\Exception\ClientException;
use Drupal\sqrl\SqrlActionPluginBase;
use Drupal\user\Entity\User;
use Drupal\user\UserInterface;
use Exception;

/**
 * Plugin implementation of the sqrl action "ident".
 *
 * @SqrlAction(
 *   id = "ident",
 *   label = @Translation("Ident"),
 *   description = @Translation("TBD.")
 * )
 */
class Ident extends SqrlActionPluginBase {

  /**
   * @var int
   */
  private $tif;

  /**
   * {@inheritdoc}
   */
  public function run(): bool {
    $this->tif = $this->client->getTif();
    switch ($this->client->getNut()->getClientOperation()) {
      case 'register':
        return $this->register();
        break;

      case 'login':
        return $this->login();
        break;

      case 'link':
        return $this->setkey();
        break;

    }
    return FALSE;
  }

  /**
   * @return bool
   * @throws \Drupal\sqrl\Exception\ClientException
   */
  private function login(): bool {
    if (!$this->user) {
      return $this->register();
    }

    if (!$this->client->getIdentity()->isEnabled()) {
      throw new ClientException('SQRL disabled for this account');
    }

    if (!$this->user->isActive()) {
      return FALSE;
    }

    if (($this->tif & $this->client::FLAG_PIDK_MATCH) === $this->client::FLAG_PIDK_MATCH) {
      if ($this->setkey()) {
        return TRUE;
      }
    }
    else {
      return TRUE;
    }
    return FALSE;
  }

  /**
   * @return bool
   * @throws \Drupal\sqrl\Exception\ClientException
   */
  private function register(): bool {
    if ($this->user) {
      return $this->login();
    }

    $registerMethod = $this->configFactory->get('user.settings')->get('register');
    if ($registerMethod === UserInterface::REGISTER_ADMINISTRATORS_ONLY) {
      // We can't create new accounts on this server.
      $this->log->debug('No user registration allowed!');
      return FALSE;
    }

    $userinfo = array(
      'name' => '',
      'pass' => $this->random->string(32),
      'init' => '',
      'status' => $registerMethod === UserInterface::REGISTER_VISITORS,
      'access' => $this->time->getRequestTime(),
    );
    $pattern = $this->configFactory->get('sqrl.settings')->get('account_name_pattern');
    if (strpos($pattern, '[rand]') === FALSE) {
      $pattern .= '[rand]';
    }
    $length = $this->configFactory->get('sqrl.settings')->get('account_name_random_length');
    while (empty($user)) {
      $name = str_replace('[rand]', $this->random->string($length), $pattern);
      if (!user_load_by_name($name)) {
        $userinfo['name'] = $userinfo['init'] = $name;
        $userinfo['mail'] = $this->dummyMail($name);
        try {
          $user = User::create($userinfo);
          $user->save();
        }
        catch (Exception $ex) {
          // Do nothing, just retry.
          unset($user);
        }
      }
    }
    $this->user = $user ?? NULL;
    if ($this->setkey()) {
      if ($registerMethod === UserInterface::REGISTER_VISITORS_ADMINISTRATIVE_APPROVAL) {
        // New accounts require admin approval.
        $this->state->addMessage($this->sqrl->getNut()->getPublicNut(), $this->t('Account created, please wait for an administrator for approval of your account.'));
      }
      return TRUE;
    }
    return FALSE;
  }

  /**
   * @return bool
   * @throws \Drupal\sqrl\Exception\ClientException
   */
  private function setkey(): bool {
    if (!$this->user) {
      return FALSE;
    }
    if ($this->identities->getIdentityByIdk($this->user->id())) {
      // Nothing changed and we have nothing to do.
      return TRUE;
    }
    if ($identity = $this->identities->getIdentityByIdk()) {
      $identity->addUser($this->user);
    }
    else {
      $identity = Identity::create([
        'idk' => $this->client->getClientVar('idk'),
        'suk' => $this->client->getClientVar('suk'),
        'vuk' => $this->client->getClientVar('vuk'),
        'user' => $this->user,
        'status' => TRUE,
      ]);
    }
    $previous_identity = NULL;
    try {
      if ($identity->save()) {
        $this->client->setResponse('suk', $this->client->getClientVar('suk'));
        $this->client->setResponse('vuk', $this->client->getClientVar('vuk'));
        if ($previous_identity = $this->identities->getIdentityByPidk()) {
          $previous_identity
            ->setSuccessor($identity)
            ->save();
        }
        return TRUE;
      }
    } catch (EntityStorageException $e) {
    }
    if ($identity !== NULL && $previous_identity !== NULL) {
      // Saving previous id failed, undo everything.
      try {
        $identity->delete();
      } catch (EntityStorageException $e) {
      }
      throw new ClientException('Updating previous ID failed');
    }
    return FALSE;
  }

  /**
   * @param string $name
   *
   * @return string
   */
  private function dummyMail($name): string {
    return urlencode($name) . Identity::DEFAULT_EMAIL_DOMAIN;
  }

}
